/************************************************************************/
/*									*/
/*  Management of bookmarks related to TOC fields.			*/
/*									*/
/************************************************************************/

#   include	"docBufConfig.h"

#   include	<stdlib.h>
#   include	<string.h>
#   include	<stdio.h>

#   include	<appDebugon.h>

#   include	<appUnit.h>

#   include	"docBuf.h"
#   include	"docTocField.h"

/************************************************************************/
/*									*/
/*  Recalculate TOC fields in a document.				*/
/*									*/
/************************************************************************/

static int docIsTocBookmark(	long *				pId,
				const DocumentField *		df )
    {
    const char *	markName;
    int			markSize;

    char *		past;
    long		id;


    if  ( docFieldGetBookmark( df, &markName, &markSize ) )
	{ LDEB(df->dfFieldNumber); return 0;	}

    if  ( markSize < 4 || strncmp( markName, "_Toc", 4 ) )
	{ return 0;	}

    id= strtod( markName+ 4, &past );
    if  ( past == markName+ 4 )
	{ return 0;	}
    if  ( past- markName < markSize )
	{ return 0;	}

    if  ( pId )
	{ *pId= id;	}
    return 1;
    }

static int docSetTocBookmark(	BufferDocument *	bd,
				DocumentField *		dfBookmark )
    {
    DocumentFieldList *	dfl= &(bd->bdFieldList);
    int			fieldNr;

    char		markName[DOCmaxBOOKMARK+1];
    int			markSize= 0;

    long		id0= 0;
    long		id1= 0;

    for ( fieldNr= 0; fieldNr < dfl->dflFieldCount; fieldNr++ )
	{
	DocumentField *		df= docGetFieldByNumber( dfl, fieldNr );
	long			id;

	if  ( ! df )
	    { continue;	}
	if  ( df->dfKind != DOCfkBOOKMARK	||
	      ! docIsTocBookmark( &id, df )	)
	    { continue;	}

	if  ( id0 <= 0 || id0 > id )
	    { id0= id;	}
	if  ( id1 <= 0 || id1 < id )
	    { id1= id;	}
	}

    if  ( id0 > 1 )
	{ sprintf( markName, "_Toc%ld", id0- 1 );	}
    else{ sprintf( markName, "_Toc%ld", id1+ 1 );	}
    markSize= strlen( markName );

    if  ( docFieldSetBookmark( dfBookmark, markName, markSize ) )
	{ SDEB(markName); return -1;	}

    return 0;
    }

void docRemoveUnbalancedTocBookmarks(	BufferDocument *	bdDoc )
    {
    DocumentFieldList *	dfl= &(bdDoc->bdFieldList);
    int			fieldNr;

    for ( fieldNr= 0; fieldNr < dfl->dflFieldCount; fieldNr++ )
	{
	DocumentField *		df= docGetFieldByNumber( dfl, fieldNr );

	if  ( df->dfSelectionScope.ssInExternalItem != DOCinBODY	)
	    { continue;	}

	if  ( df->dfKind == DOCfkBOOKMARK				&&
	      df->dfHeadPosition.epParaNr !=
				      df->dfTailPosition.epParaNr	&&
	      docIsTocBookmark( (long *)0, df )				)
	    {
	    DocumentSelection	dsInside;
	    DocumentSelection	dsAround;
	    int			headPart;
	    int			tailPart;

	    if  ( docDelimitFieldInDoc( &dsInside, &dsAround,
					    &headPart, &tailPart, bdDoc, df ) )
		{ LDEB(fieldNr); continue; }

	    if  ( docDeleteField( dsInside.dsHead.dpBi, dsInside.dsTail.dpBi,
						&headPart, &tailPart,
						headPart, tailPart,
						&(bdDoc->bdBody), bdDoc, df ) )
		{ LDEB(fieldNr);	}
	    }
	}
    }

/************************************************************************/
/*									*/
/*  Set a TOC bookmark that contains the whole paragraph.		*/
/*									*/
/************************************************************************/

int docSetParaTocBookmark(	BufferDocument *	bd,
				DocumentTree *		ei,
				BufferItem *		paraBi )
    {
    DocumentPosition	dpHead;
    DocumentPosition	dpTail;

    DocumentField *	dfBookmark= (DocumentField *)0;

    if  ( docFirstPosition( &dpHead, paraBi ) )
	{ LDEB(1); return -1;	}
    docAvoidParaHeadField( &dpHead, (int *)0, bd );

    if  ( docLastPosition( &dpTail, paraBi ) )
	{ LDEB(1); return -1;	}

    dfBookmark= docFindTypedFieldForPosition( bd, &dpHead, DOCfkBOOKMARK );
    if  ( ! dfBookmark )
	{
	DocumentSelection	dsBookmark;
	DocumentSelection	dsInsideMark;
	DocumentSelection	dsAroundMark;
	int			partHead= -1;
	int			partTail= -1;

	TextAttribute	taSet;
	PropertyMask	taSetMask;

	utilInitTextAttribute( &taSet );
	PROPmaskCLEAR( &taSetMask );

	if  ( dpTail.dpStroff != dpHead.dpStroff )
	    {
	    const int	singlePara= 1;

	    docSetRangeSelection( &dsBookmark, &dpHead, &dpTail, 1, -1, -1 );

	    if  ( docSurroundTextSelectionByField( &dfBookmark,
						&dsInsideMark, &dsAroundMark,
						&partHead, &partTail, bd, ei,
						singlePara, &dsBookmark, 
						&taSetMask, &taSet ) )
		{ LDEB(1); return -1;	}

	    if  ( docSetTocBookmark( bd, dfBookmark ) )
		{ LDEB(1); return -1;	}
	    }
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  Make sure that all leaves in the outlinelevel tree have a bookmark	*/
/*  that can be used in the table of contents.				*/
/*									*/
/************************************************************************/

static int docSetTocBookmarksForLevel(	BufferDocument *		bd,
					DocumentTree *			ei,
					const ListNumberTreeNode *	lntn )
    {
    int		i;
    int		rval= 0;

    if  ( lntn->lntnIsLeaf )
	{
	BufferItem *	paraBi;

	paraBi= docGetParagraphByNumber( ei, lntn->lntnParaNr );
	if  ( ! paraBi )
	    { XDEB(paraBi); rval= -1;	}
	else{
	    if  ( docSetParaTocBookmark( bd, ei, paraBi ) )
		{ LDEB(lntn->lntnParaNr); rval= -1;	}
	    }
	}

    for ( i= 0; i < lntn->lntnChildCount; i++ )
	{
	if  ( docSetTocBookmarksForLevel( bd, ei, &(lntn->lntnChildren[i]) ) )
	    { LDEB(i); rval= -1;	}
	}

    return rval;
    }

int docSetTocBookmarks(	BufferDocument *		bd )
    {
    DocumentTree *	ei= &(bd->bdBody);

    DocumentFieldList *	dfl= &(bd->bdFieldList);
    int			fieldNr;

    if  ( docSetTocBookmarksForLevel( bd, ei, &(ei->eiOutlineTree) ) )
	{ LDEB(1); return -1;	}

    for ( fieldNr= 0; fieldNr < dfl->dflFieldCount; fieldNr++ )
	{
	DocumentField *		dfTc= docGetFieldByNumber( dfl, fieldNr );
	DocumentField *		dfBookmark;
	DocumentSelection	dsInsideTc;
	DocumentSelection	dsAroundTc;
	DocumentSelection	dsInsideMark;
	DocumentSelection	dsAroundMark;
	int			headPart;
	int			tailPart;

	if  ( ! dfTc )
	    { continue;	}
	if  (  ( dfTc->dfKind != DOCfkTC && dfTc->dfKind != DOCfkTCN )	||
	      dfTc->dfSelectionScope.ssInExternalItem != DOCinBODY	)
	    { continue;	}

	if  ( docDelimitFieldInDoc( &dsInsideTc, &dsAroundTc,
					    &headPart, &tailPart, bd, dfTc ) )
	    { LDEB(1); return -1;	}

	dfBookmark= docFindTypedFieldForPosition( bd, &(dsInsideTc.dsHead),
							    DOCfkBOOKMARK );
	if  ( ! dfBookmark )
	    {
	    TextAttribute	taSet;
	    PropertyMask	taSetMask;
	    const int		singlePara= 1;

	    utilInitTextAttribute( &taSet );
	    PROPmaskCLEAR( &taSetMask );

	    if  ( docSurroundTextSelectionByField( &dfBookmark,
						&dsInsideMark, &dsAroundMark,
						&headPart, &tailPart, bd, ei,
						singlePara, &dsAroundTc, 
						&taSetMask, &taSet ) )
		{ LDEB(1); return -1;	}

	    if  ( docSetTocBookmark( bd, dfBookmark ) )
		{ LDEB(1); return -1;	}
	    }
	}

    return 0;
    }

